#include <LiquidCrystal.h>
#include <Wire.h>

/*----------------------------------------------------------------------------
  I2C Audio Controller

  Controls I2C pre amp boards designed by Rory Thompson
  hackaday.io/roryt

  Each board has 4 stereo inputs. These are switched as desired to 2 bi amped
  sterio outputs (zones) with volume control. Up to 8 boards can be connected
  together to form a 16 zone system. The system is designed to feed cheep
  class D amplifier modules from ebay. The standby and mute functions of the
  amplifiers are also controlled by the I2C interface

  Changelog:

  3.0 - 03/02/19
  Combined test functions together

  ----------------------------------------------------------------------------*/



// channel and vol selects
byte boardCh[16]      = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; // board channel select. Address 0 output 1, address 0 output 2, address 1 output 0... etc
byte boardChLast[16]  = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1}; // holds last changed value
byte boardVol[16]     = {33, 33, 33, 33, 33, 33, 33, 33, 33, 33, 33, 33, 33, 33, 33, 33}; // board volume select
byte boardVolLast[16] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1}; // holds last changed value
byte boardMute[16]    = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1}; // holds amp mutes

// Min & Max Values for board arrays - Values are limited to these
// not implemented

unsigned long taskList[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; // each register contains bits detailing commands to be sent to I2C digital pots. See functionTask in functions
boolean taskListFull = LOW; // this goes high if the task list gets full

// Default I2C IO masks for initial setup
byte portADefault = B01100110; // all channels off
byte portBDefault = 0xFF;      // all pins high

// Buttons & navigation - Buttons go low when pressed
int buttons[4] {5, 4, 7, 6}; // left button, right button, up button, down button
byte buttonState = B00001111;
int horizMenPos = 0;
int vertMenPos = 0;

// globals for screen
const int rs = 8, en = 9, d4 = 15, d5 = 14, d6 = 16, d7 = 10;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);


// Values for timer
unsigned long lastMillis = 0;
unsigned long timer = 1;   //timer fire time
int taskCount = 0;
boolean task1Flag = LOW;
int taskCount2 = 0;
boolean task2Flag = LOW;

//----------------------------------------------------------------------------
// Setup
//----------------------------------------------------------------------------


void setup() {

  // Begin Serial
  Serial.begin(9600);

  // Begin I2C
  Wire.begin();

  // Begin LCD
  lcd.begin(16, 2);

  // Setup Buttons

  for (int i = 0; i <= 3; i++) {             // set all buttons to inputs
    pinMode(buttons[i], INPUT);
  }
  pinMode(21, OUTPUT);


  for (int i = 0; i <= 7; i++) {             // update all boards IO
    functionSendIO(i, 0x00, 0);              // sends IODIRA val 0 to all addresses to configure as outputs
    functionSendIO(i, 0x01, 0);              // sends IODIRB val 0 to all addresses to configure as outputs
    functionSendIO(i, 0x14, portADefault);   // sends OLATA to defaults
    functionSendIO(i, 0x15, portBDefault);   // sends OLATB to defaults
  }
  /*
    for (int i = 0; i <= 7; i++) {
      functionSendPot(i, 33, 33);              // mute all amps and set all pots to mute
    }

  */
}

//----------------------------------------------------------------------------
// Loop
//----------------------------------------------------------------------------



void loop() {


  if ((millis() - lastMillis) > timer) {                          // do every 1ms
    lastMillis = millis();
    taskCount ++;
    taskCount2 ++;
    if (taskCount > 50) {                                        // do every 100ms set to 50 because clock seems to be halved?

      taskCount = 0;
      task1Flag = HIGH;
    }
    if (taskCount2 > 10) {                                        // do every 20ms set to 10 because clock seems to be halved?

      taskCount2 = 0;
      task2Flag = HIGH;
    }

    buttonState = functionCheckButtons(buttonState, buttons);
    if ((buttonState & B10001000) == 128) { // left button has been pressed for longer then 2ms
      vertMenPos -- ;
      vertMenPos = functionLimits(vertMenPos, 0, 32);
    }
    if ((buttonState & B01000100) == 64) { // right button has been pressed for longer then 2ms
      vertMenPos ++ ;
      vertMenPos = functionLimits(vertMenPos, 0, 32);
    }
    if ((buttonState & B00100010) == 32) { // up button has been pressed for longer then 2ms

      if (vertMenPos <= 15) {
        boardCh[vertMenPos] ++ ;
        boardCh[vertMenPos] = functionLimits(boardCh[vertMenPos], 0, 4);
      }
      if (vertMenPos > 16) {
        if (boardVol[vertMenPos - 17] < 33) {
          boardVol[vertMenPos - 17] ++ ;
        }
      }
    }

    if ((buttonState & B00010001) == 16) { // down button has been pressed for longer then 2ms
      if (vertMenPos <= 15) {
        boardCh[vertMenPos] -- ;
        boardCh[vertMenPos] = functionLimits(boardCh[vertMenPos], 0, 4);
      }
      if (vertMenPos > 16) {
        if (boardVol[vertMenPos - 17] != 0) {
          boardVol[vertMenPos - 17] -- ;
        }
      }
    }
  }

  if (taskCount >= 1 && task1Flag == HIGH) {                        // do every 100ms
    digitalWrite(21, HIGH);
    functionUpdateDisplay();                                        // update display
    digitalWrite(21, LOW);
    task1Flag = LOW;                                                // dont do it again
  }


  if (taskCount2 >= 1 && task2Flag == HIGH) {                        // do every 20ms

    //digitalWrite(21, HIGH);
    if (taskList[0] != 0) {                                          // if task list has value
      taskList[0] = functionSendPot(taskList[0]);                    // send to pot
    }
    if (taskList[0] == 0 && taskList[1] != 0) {                       // shift data left if task has been completed
      for (int i = 0; i <= 15 ; i++) {
        taskList[i] = taskList[i + 1];
      }
    }

    task2Flag = LOW;                                                 // dont do it again
    // digitalWrite(21, LOW);
  }









  for (int i = 0; i <= 7 ; i++) {            // check if any channels have been changed and run function if so
    if ((boardCh[(i * 2)] != boardChLast[(i * 2)]) || (boardCh[(i * 2) + 1] != boardChLast[(i * 2 + 1)])) {
      functionSendCh(i, boardCh[(i * 2)], boardCh[(i * 2 + 1)]);
      boardChLast[(i * 2)] = boardCh[(i * 2)] ;        // update ChLast register
      boardChLast[(i * 2 + 1)] = boardCh[(i * 2) + 1];
    }

  }

  for (int i = 0; i <= 15 ; i++) {            // check if any channels are zero
    if (boardCh[i] == 0) {
      boardVol[i] = 33;                       // change volume to 33 if so (mute in digital pot)
      boardMute[i] = 1;                       // udate list of mute. will put amp into standby
    }
    else {
      boardMute[i] = 0;
    }
  }


  for (int i = 0; i <= 15 ; i++) {            // check if any volumes have been changed
    if (boardVol[i] != boardVolLast[i]) {
      for (byte j = 0; j <= 15 ; j++) {      // scan array for next available blank space and add task
        if (taskList[j] == 0) {
          taskList[j] = functionTask( (i / 2), (i & 1), boardMute[((i / 2) * 2)], boardMute[((i / 2) * 2) + 1], boardVol[i], boardVol[i], boardVol[i], boardVol[i]);
          boardVolLast[i] = boardVol[i];
          break;
        }
        if (j >= 15) {
          taskListFull = HIGH;               // task list is full
        }
      }
    }
  }

}

//----------------------------------------------------------------------------
// Functions
//----------------------------------------------------------------------------


// This function arranges bits which is later fed into functionSendPot
// returns 32 bit value.
// bits 31-30 = task counter. This starts as 3 (B11) functionSendPot runs 3 times to send message, each run decrements this counter.
// bits 29-27 = board address (input as byte a)
// bit 26 = amp number. 1 = amp 2, 0 = amp 1 (input as byte b)
// bit 25 = amp 0 mute (1 = mute) input c
// bit 24 = amp 1 mute (1 = mute) input d
// bit 23-19 = pot 0 on high frequency output (right)  (input as byte pot0HF)
// bit 17-12 = pot 1 on high frequency output (left)  (input as byte pot1HF)
// bit 11-6 = pot 0 on low frequency output (right)  (input as byte pot0LF)
// bit 5-0 = pot 1 on low frequency output (left)  (input as byte pot1LF)

unsigned long functionTask(byte a, byte b, byte c, byte d, byte pot0HF, byte pot1HF, byte pot0LF, byte pot1LF) {
  unsigned long val = 0;

  val = 3; //  add task counter bits
  val = val << 3; // shift bits left 3 to make room for board address
  val = val + (a & 7); // add board address. $ with 7 to ensure board address isnt more than 3 bits long
  val = val << 1; // shift bits left 1 to make room for amp numer
  val = val + (b & 1); // add board address. $ with 1 to ensure amp number isnt more than 1 bits long
  val = val << 1; // shift bits left 6 to make room for amp 0 mute value
  val = val + (c & 1); // add amp 0 mute. $ with 1 to ensure number isnt more than 1 bits long
  val = val << 1; // shift bits left 6 to make room for amp 0 mute value
  val = val + (d & 1); // add amp 0 mute. $ with 1 to ensure number isnt more than 1 bits long
  val = val << 6; // shift bits left 6 to make room for pot value
  val = val + (pot0HF & 63); // add board address. $ with 63 to ensure value isnt more than 6 bits long
  val = val << 6; // shift bits left 6 to make room for pot value
  val = val + (pot1HF & 63); // add board address. $ with 63 to ensure value isnt more than 6 bits long
  val = val << 6; // shift bits left 6 to make room for pot value
  val = val + (pot0LF & 63); // add board address. $ with 63 to ensure value isnt more than 6 bits long
  val = val << 6; // shift bits left 6 to make room for pot value
  val = val + (pot1LF & 63); // add board address. $ with 63 to ensure value isnt more than 6 bits long

  return val;
}

// This function sends values to digital pots as set via functionTask
// should be run three times (untill returns zero)
// will retun original vaue with task counter updated on each iteration.
// updates digital IO to send to pots
// will also mute amplifiers on selected output if any pot vales are 33
// inout bits as follows:
// bits 31-30 = task counter. This starts as 3 (B11) functionSendPot runs 3 times to send message, each run decrements this counter.
// bits 29-27 = board address (input as byte a)
// bit 26 = amp number. 1 = amp 2, 0 = amp 1 (input as byte b)
// bit 25 = amp 0 mute (1 = mute) input c
// bit 24 = amp 1 mute (1 = mute) input d
// bit 23-19 = pot 0 on high frequency output (right)  (input as byte pot0HF)
// bit 17-12 = pot 1 on high frequency output (left)  (input as byte pot1HF)
// bit 11-6 = pot 0 on low frequency output (right)  (input as byte pot0LF)
// bit 5-0 = pot 1 on low frequency output (left)  (input as byte pot1LF)

unsigned long functionSendPot(unsigned long val) {
  ; //this gets called every 20ms when required
  unsigned long val2 = val; //copy val (val is modified later)
  byte pot1LF = (val2 & 63);
  val2 = val2 >> 6;
  byte pot0LF = (val2 & 63);
  val2 = val2 >> 6;
  byte pot1HF = (val2 & 63);
  val2 = val2 >> 6;
  byte pot0HF = (val2 & 63);
  val2 = val2 >> 6;
  byte d = (val2 & 1);  //amp 1 mute
  val2 = val2 >> 1;
  byte c = (val2 & 1);  //amp 0 mute
  val2 = val2 >> 1;
  byte b = (val2 & 1);  //amp number
  val2 = val2 >> 1;
  byte a = (val2 & 7);  //board address
  val2 = val2 >> 3;
  byte task = (val2 & 3);  //task count

  a = a + 0x20; // add MCP23017 address offset

  byte bits = 0; // bits for IO, all 0

  if (c == 1) {              // if amp 0 requires muting
    bits = bits | 192;       // set bits 7, 6
  }
  if (d == 1) {              // if amp 1 requires muting
    bits = bits | 48;        // set bits 5, 4
  }

  if (b == 0) {              // if command is for amp 0 set starting bits for IO
    bits = bits | 7;         // bits 2,1,0 set
  }
  else {                     // if command is for amp 1 set starting bits for IO
    bits = bits | 13;        // bits 3,2,0 set
  }



  pot1LF |= (1 << 6); // sets bit 6 to be 1 as required by pot
  pot1HF |= (1 << 6); // sets bit 6 to be 1 as required by pot


  if (task == 3) { // first time this task has been executed. Updates CE for HF pot

    Wire.beginTransmission(a);  // send to board address
    Wire.write(byte(0x15));     // sends instruction byte
    Wire.write(byte(bits));     // pulls CS low
    Wire.endTransmission();     //

    val &= ~((unsigned long)1 << 30);  // clear bit 30 making task counter of val = 2
    return val;
  }


  if (task == 2) { // second time this task has been executed. Sends HF pot & updates CE for pot 2

    Wire.beginTransmission(0x2C);  // write to pot HF. Comms is now enabled from IO bits set above
    Wire.write(pot0HF);                 //
    Wire.write(pot1HF);                 //
    Wire.endTransmission();        //

    bits = (bits & ~B00001111 | ((bits >> 1) | 8) & B00001111);  // shift lower 4 bits and make bit 3 1

    Wire.beginTransmission(a);  // send to board address
    Wire.write(byte(0x15));     // sends instruction byte
    Wire.write(bits);     // sets CS
    Wire.endTransmission();     //

    val &= ~((unsigned long)1 << 31);  // clear bit 31 making task counter of val = 0
    val |= ((unsigned long)1 << 30);  // set bit 30 making task counter of val = 1
    return val;
  }

  if (task == 1) { // third time this task has been executed. Sends LF pot & sets all IO port bits to 1


    Wire.beginTransmission(0x2C);  // write to pot LF. Comms is now enabled from IO bits set above
    Wire.write(pot0LF);                 //
    Wire.write(pot1LF);                 //
    Wire.endTransmission();        //

    bits = (bits & ~B00001111 | 15); // make lower 4 bits 1

    Wire.beginTransmission(a);  // send to board address
    Wire.write(byte(0x15));     // sends instruction byte
    Wire.write(bits);     // sets CS
    Wire.endTransmission();     //

    return 0; // this task has been completed so return zero
  }
}

// this sends to IO portA to set channels
int functionSendCh(int x, int y, int z) {     // x = address, y = output 1 channel (0 - 4), y = output 2 channel (0 - 4)

  x = x + 0x20;                 // add MCP23017 address offset
  byte mask = B00001111;
  byte portA;


  if (y == 0) {                             //configure port a bits bassed on chanel switching requirement
    portA = (portA & ~mask) | (6 & mask);
  }
  if (y == 1) {
    portA = (portA & ~mask) | (7 & mask);
  }
  if (y == 2) {
    portA = (portA & ~mask) | (4 & mask);
  }
  if (y == 3) {
    portA = (portA & ~mask) | (2 & mask);
  }
  if (y == 4) {
    portA = (portA & ~mask) | (14 & mask);
  }
  if (z == 0) {
    portA = (portA & mask) | (96 & ~mask);
  }
  if (z == 1) {
    portA = (portA & mask) | (112 & ~mask);
  }
  if (z == 2) {
    portA = (portA & mask) | (64 & ~mask);
  }
  if (z == 3) {
    portA = (portA & mask) | (32 & ~mask);
  }
  if (z == 4) {
    portA = (portA & mask) | (224 & ~mask);
  }

  Wire.beginTransmission(x);  // write port A bits
  Wire.write(byte(0x14));     //
  Wire.write(portA);  //
  Wire.endTransmission();     //

}


// general send I2C
int functionSendIO(byte x, byte y, byte z) {   // x = address, y = instruction byte, z = value
  x = x + 0x20;                 // add MCP23017 address offset
  Wire.beginTransmission(x); // transmit to device
  Wire.write(y);       // sends instruction byte
  Wire.write(z);        // sends  value byte
  Wire.endTransmission();       // stop transmitting
}


// Updates display
int functionUpdateDisplay() {
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("AMP");


  if (vertMenPos <= 15) {
    lcd.setCursor(4, 0);
    lcd.print(vertMenPos);
    lcd.setCursor(8, 0);
    lcd.print("INPUT");
    lcd.setCursor(8, 1);
    lcd.print("CHANNEL");
    lcd.setCursor(0, 1);
    lcd.print(boardCh[vertMenPos]);

  }

  if (vertMenPos > 16) {
    lcd.setCursor(4, 0);
    lcd.print(vertMenPos - 17);
    lcd.setCursor(8, 0);
    lcd.print("OUTPUT");
    lcd.setCursor(8, 1);
    lcd.print("dB ATEN");
    lcd.setCursor(0, 1);
    lcd.print(boardVol[vertMenPos - 17]);
  }

  /*
    for (int i = 0; i <= 15; i++) {
      Serial.print(boardVol[i]);
    }

    Serial.println();
  */
}



// Checks Buttons
byte functionCheckButtons(byte x, int z[4]) { // byte X 7 - 4 are last button states, byre 3 - 0 are current button states, array of buttons

  int y[4] {5, 4, 7, 6};

  for (int i = 0; i <= 3; i++) {
    x = x << 1;
    x = x + digitalRead(y[i]);
  }
  return x;
}

// returns an input clipped between high and low limits
int functionLimits(int x, int y, int z) {   // x = input, y = low limit, z = high limit

  if ( x > z) {
    x = z;
  }

  if ( x < y) {
    x = y;
  }
  return x;
}
